package com.ae7q.communication;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.util.logging.*;
import java.util.prefs.*;
import javax.swing.*;
//import javax.swing.event.*;
// import javax.comm.*;
import purejavacomm.*;              // Uses com.sun.jna.*

import com.ultimeth.util.*;

//==================================================================================================
/** DstarTerminal is a D-Star terminal program.
 *
 *  @version    2006.04.09
 *  @author     Copyright © 2006 by Dean K. Gibson.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
@SuppressWarnings( "serial" )
class DstarTerminal extends Box {
    private static final String        TITLE             = "D-Star terminal";
    private static final String        COPYRIGHT         = "Copyright \u00A9 2006 by Dean K. Gibson";
    private static final char          endOfLine         = '\r';
    private static final JFrame        jFrame            = new JFrame( TITLE );
    private static final JTabbedPane   jTabbedPane       = new JTabbedPane();
    private static final Map <String, DstarTerminal>
                                       hashMap           = new HashMap <String, DstarTerminal>();
    private static final AbstractList <Component>
                                       uiComponents      = new ArrayList <Component>();
    private static final Preferences   preferences       = Preferences
                                                           .userNodeForPackage( DstarTerminal.class );
    private static final DstarTerminal unproto           = new DstarTerminal();
    private static       String             localCall;
    private static       CommPortIdentifier serialPortId = null;
    private static       int                baudRate     = 9600;
    private static       DstarSerialIO      dstarSerial;

           private final MouseAdapter  mouseAdapter     = new MouseAdapter(){
                        public void mousePressed( MouseEvent event ){
                            if ( event.isPopupTrigger() ){      // for Solaris
                                Tools.showMenu( jPopupMenu, event );
                            }
                        }
                        public void mouseReleased( MouseEvent event ){
                            if ( event.isPopupTrigger() ){      // for Windows
                                Tools.showMenu( jPopupMenu, event );
                            }
                        }
                    };// MouseAdapter subclass
           private final JPopupMenu    jPopupMenu       = new JPopupMenu();
           private final JTextField    messageField     = new JTextField( 100 );
           private final Logger        log              = Log.get( this.getClass() );
           private final LogPane       handler          = new LogPane( new JTextArea() );
           private final String        remoteCall;
           private       Font          textFont;

    {
        log.addHandler( handler );
        add( new JScrollPane( handler.pane ) );
        handler.pane.setLineWrap( true );
        handler.pane.setWrapStyleWord( true );
        handler.pane.addMouseListener( mouseAdapter );
        initMenus( this );
    }

    //----------------------------------------------------------------------------------------------
    /** Application main entry point.  Create the main window and call common program setup.
     *
     *  @param  args                    an array of command line arguments.
     */
    @SuppressWarnings( {"unchecked", "fallthrough"} )
    public static void main( String[] args ){
        StringTokenizer tokens;

        jFrame.getContentPane().add( jTabbedPane );
        unproto.handler.setLevel( Level.ALL );
        switch ( args.length ){
        case 2:
            for ( CommPortIdentifier commPortId
                  : new ToIterable <CommPortIdentifier>(
                                   CommPortIdentifier.getPortIdentifiers() ) ){  // [unchecked]
                if ( commPortId.getPortType() == CommPortIdentifier.PORT_SERIAL ){
                    if ( commPortId.getName().equalsIgnoreCase( args[ 1 ] ) ){
                        serialPortId = commPortId;
                        break;
                    }
                }
            }
            if ( serialPortId == null ){
                unproto.log.severe( "*** Serial port " + args[ 1 ] + " not found ! ***" );
            }
        case 1:  // [fallthrough]
            localCall = args[ 0 ];
            break;
        case 0:
            unproto.log.severe( "*** Callsign and serial port not provided ! ***" );
            break;
        }
        unproto.log.severe( "Continuing with " + localCall + "/" + args[ 1 ] );
        initMenus();
        jFrame.setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );
        jFrame.setBounds( 200, 10, 800, 720 );
        jFrame.setVisible( true );
        new Thread(){
            public void run(){
                listen();
            }
        }.start();
    } // DstarTerminal.main

    //----------------------------------------------------------------------------------------------
    /** Initialize window menu items.
     */
    private static void initMenus(){
        JMenuBar             jMenuBar = new JMenuBar();
        JMenu                jMenu,
                             jSubMenu;
        JMenuItem            jMenuItem;
        JRadioButtonMenuItem button;
        ButtonGroup          buttonGroup;
        String               value,
                             currLnfName = UIManager.getLookAndFeel().getName();

        uiComponents.add( jFrame );
        jFrame.setJMenuBar( jMenuBar );

        jMenu = new JMenu( "File" );
        jMenu.setMnemonic( KeyEvent.VK_F );
        jMenuBar.add( jMenu );

        jMenuItem = new JMenuItem( "Connect ..." );
        jMenuItem.setMnemonic( KeyEvent.VK_C );
        jMenuItem.addActionListener( new ActionListener(){
                public void actionPerformed( ActionEvent event ){
                    final JDialog    jDialog      = new JDialog( jFrame, "Connection Parameters", true );
                    final JTextField stationField = new JTextField( DstarSerialIO.Segment.ADDRESS_SIZE );
                          JButton    jButton      = new JButton( "Connect" );
                          Box        hBox;
                          Container  container    = jDialog.getContentPane();
                    container.setLayout( new BoxLayout( container, BoxLayout.Y_AXIS ) );
                    container.add( new JPanel() );

                    hBox = new Box( BoxLayout.X_AXIS );
                    hBox.add( new JPanel() );
                    hBox.add( new JLabel( "Destination station: " ) );
                    hBox.add( stationField );
                    hBox.add( new JPanel() );
                    container.add( hBox );

                    container.add( new JPanel() );
                    
                    hBox = new Box( BoxLayout.X_AXIS );
                    hBox.add( new JPanel() );
                    hBox.add( jButton );
                    hBox.add( new JPanel() );
                    container.add( hBox );

                    container.add( new JPanel() );
                    stationField.setMaximumSize( stationField.getPreferredSize() );

                    ActionListener connectListener = new ActionListener(){
                        public void actionPerformed( ActionEvent event ){
                            String  remoteCall = stationField.getText().toUpperCase();
                            try {
                                dstarSerial.send( new DstarSerialIO.Segment( unproto.log, localCall, remoteCall, null ) );
                            } catch ( Exception exception ){
                                unproto.log.log( Level.SEVERE,
                                                 "Communications error sending a message",
                                                 exception );
                            }
                            jDialog.dispose();
                        }
                    }; // ActionListener implementation
                    jButton.addActionListener( connectListener );
                    stationField.addActionListener( connectListener );
                    jDialog.pack();
                    jDialog.setLocation( - jDialog.getWidth()  / 2,
                                         - jDialog.getHeight() / 2 );
                    jDialog.setLocationRelativeTo( jFrame );
                    jDialog.setVisible( true );
                }
            } ); // ActionListener implementation
        jMenu.add( jMenuItem );

        jMenu.addSeparator();

        jMenuItem = new JMenuItem( "Exit" );
        jMenuItem.setMnemonic( KeyEvent.VK_E );
        jMenuItem.addActionListener( new ActionListener(){
                public void actionPerformed( ActionEvent event ){
                    switch ( JOptionPane.showConfirmDialog( jFrame, "Do you really want to exit?",
                                                            "Exit", JOptionPane.OK_CANCEL_OPTION,
                                                                    JOptionPane.QUESTION_MESSAGE ) ){
                    case JOptionPane.OK_OPTION:
                        System.exit( 0 );
                    }
                }
            } ); // ActionListener implementation
        jMenu.add( jMenuItem );

        jMenu = new JMenu( "Window" ); 
        jMenu.setMnemonic( KeyEvent.VK_W );
        jMenu.addActionListener( new ActionListener(){
                public void actionPerformed( ActionEvent event ){
                    Component component = jTabbedPane.getSelectedComponent();
                    if ( component instanceof DstarTerminal ){
                        DstarTerminal terminal = (DstarTerminal) component;
                        terminal.jPopupMenu.show( component, component.getWidth()  / 2,
                                                             component.getHeight() / 2 );
                    }
                }
            } ); // ActionListener implementation
        jMenuBar.add( jMenu );

        jMenu = new JMenu( "Options" );
        jMenu.setMnemonic( KeyEvent.VK_O );
        jMenuBar.add( jMenu );

        // Set up Look & Feel submenu
        value = preferences.get( "Window.lookAndFeel", "Native" );
        jSubMenu = new JMenu( "Look & Feel" );
        buttonGroup = new ButtonGroup();
        jMenuItem = null;
        for ( UIManager.LookAndFeelInfo lookAndFeelInfo : Tools.lookAndFeelInfos ){
            final String    className   = lookAndFeelInfo.getClassName();
            final String    name        = lookAndFeelInfo.getName();
                  int       key         = Character.toUpperCase( name.charAt( 0 ) );
            button = new JRadioButtonMenuItem( name, name.equals( currLnfName ) );
            button.setToolTipText( "Set the Look & Feel to " + name );
            button.addActionListener( new ActionListener(){
                    public void actionPerformed( ActionEvent event ){
                        try {
                            UIManager.setLookAndFeel( className );
                            unproto.log.config( "Setting: "
                                              + UIManager.getLookAndFeel().getDescription() );

                            for ( Component component : uiComponents ) {
                                SwingUtilities.updateComponentTreeUI( component );
                            }
                            preferences.put( "Window.lookAndFeel", name );
                        } catch ( Exception exception ){  // multiple
                            unproto.log.log( Level.WARNING, "Error setting the Look & Feel",
                                             exception );
                        }
                    }
                } ); // ActionListener implementation
            buttonGroup.add( button );
            jSubMenu.add( button );

            if ( name.equalsIgnoreCase( value )
                    || "Native".equalsIgnoreCase( value )
                        &&  className.equals( UIManager.getSystemLookAndFeelClassName() ) ){
                jMenuItem = button;
            }

        }
        jMenu.add( jSubMenu );
        if ( jMenuItem != null ){
            jMenuItem.doClick();
        } else {
            unproto.log.warning( "*** '" + value
                               + "' initial look and feel not found, using default ***" );
        }

        jMenu = new JMenu( "Help" );
        jMenu.setMnemonic( KeyEvent.VK_H );
        jMenuBar.add( jMenu );

        jMenuItem = new JMenuItem( "About" );
        jMenuItem.setMnemonic( KeyEvent.VK_A );
        jMenuItem.addActionListener( new ActionListener(){
                public void actionPerformed( ActionEvent event ){
                    JOptionPane.showMessageDialog( jFrame,  TITLE + Log.endOfLine + COPYRIGHT,
                                                   "About", JOptionPane.INFORMATION_MESSAGE );
                }
            } ); // ActionListener implementation
        jMenu.add( jMenuItem );
    } // DstarTerminal.initMenus

    //----------------------------------------------------------------------------------------------
    /** Initialize window menu items.
     *
     *  @param  terminal                the terminal window Container.
     */
    private void initMenus( final DstarTerminal terminal ){
        JMenu                jSubMenu;
        JMenuItem            jMenuItem;
        JRadioButtonMenuItem button;
        ButtonGroup          buttonGroup;
        String               value;

        uiComponents.add( jPopupMenu );

        jMenuItem = new JMenuItem( "Disconnect" );
        jMenuItem.setMnemonic( KeyEvent.VK_D );
        jMenuItem.addActionListener( new ActionListener(){
                public void actionPerformed( ActionEvent event ){
                    Component component = jTabbedPane.getSelectedComponent();
                    if ( terminal.remoteCall != null ){
                        try {
                            dstarSerial.send( new DstarSerialIO
                                                  .Segment( log, localCall, terminal.remoteCall, null ) );
                        } catch ( Exception exception ){
                            unproto.log.log( Level.SEVERE, "Communications error sending a message",
                                             exception );
                        }
                        return;
                    }
                    JOptionPane
                     .showMessageDialog( jFrame, "This window does not have an active connection !",
                                                 "Disconnect", JOptionPane.ERROR_MESSAGE );
                }
            } ); // ActionListener implementation
        jPopupMenu.add( jMenuItem );

        jMenuItem = new JMenuItem( "Clear window" );
        jMenuItem.setMnemonic( KeyEvent.VK_C );
        jMenuItem.addActionListener( new ActionListener(){
                public void actionPerformed( ActionEvent event ){
                    terminal.handler.clear();
                }
            } ); // ActionListener implementation
        jPopupMenu.add( jMenuItem );

        jMenuItem = new JMenuItem( "Delete window" );
        jMenuItem.setMnemonic( KeyEvent.VK_D );
        jMenuItem.addActionListener( new ActionListener(){
                public void actionPerformed( ActionEvent event ){
                    String    text      = "Non-connection windows cannot be deleted !";
                    Component component = jTabbedPane.getSelectedComponent();
                    if ( terminal.remoteCall != null ){
                        text = "Windows with active connections cannot be deleted !";
                        if ( ! terminal.messageField.isEditable() ){
                            jTabbedPane.remove( component );
                            return;
                        }
                    }
                    JOptionPane.showMessageDialog( jFrame, text, "Tabbed Window Delete",
                                                   JOptionPane.ERROR_MESSAGE );
                }
            } ); // ActionListener implementation
        jPopupMenu.add( jMenuItem );

        // Set up Font menu item
        jMenuItem = new JMenuItem( "Select Font" );
        jMenuItem.setToolTipText( "Select the default text font" );
        jMenuItem.addActionListener( new ActionListener(){
                public void actionPerformed( ActionEvent event ){
                    Component component = jTabbedPane.getSelectedComponent();
                    JFontChooser jFontChooser = new JFontChooser( jFrame, "Select a font",
                                                                  terminal.textFont );
                    jFontChooser.setVisible( true );
                    Font font = jFontChooser.getSelectedFont();
                    if ( font != null ){
                        terminal.textFont = font;
                        terminal.handler.pane.setFont( font );
                        if ( terminal == unproto ){
                            preferences.put(    "Window.font.name",  font.getName() );
                            preferences.putInt( "Window.font.style", font.getStyle() );
                            preferences.putInt( "Window.font.size",  font.getSize() );
                        }
                        terminal.log.config( "Setting: font '" + font.getName() + "', "
                                                               + font.getSize() + "pt" );
                    }
                }
            } ); // ActionListener implementation
        jPopupMenu.add( jMenuItem );

        // Set up Log Options submenu
        String currentName = (unproto != null ? unproto : terminal).handler.getLevel().getName();
        value = preferences.get( "Log.level", currentName );
        jSubMenu = new JMenu( "Logging level" );
        buttonGroup = new ButtonGroup();
        jMenuItem = null;
        for ( final Level level : LogPane.levels ){
            final String name = level.getName();
            button = new JRadioButtonMenuItem( Tools.initialCase( name ),
                                               name.equals( currentName ) );
            button.setToolTipText( "Set the " + name + " level of activity logging" );
            button.addActionListener( new ActionListener(){
                    public void actionPerformed( ActionEvent event ){
                        Component component = jTabbedPane.getSelectedComponent();
                        if ( component instanceof DstarTerminal ){
                            DstarTerminal terminal = (DstarTerminal) component;
                            terminal.handler.setLevel( level );
                            if ( terminal == unproto ){
                                preferences.put( "Log.level", name );
                            }
                            terminal.log.config( "Setting: The " + name
                                               + " level of activity logging" );
                        }
                    }
                } ); // ActionListener implementation
//          keyStrokeMap.put( KeyStroke.getKeyStroke( logLevel.charAt( 0 ) ), button );
            buttonGroup.add( button );
            jSubMenu.add( button );
            if ( name.equalsIgnoreCase( value ) ){
                jMenuItem = button;
            }
        }
        jPopupMenu.add( jSubMenu );
        if ( jMenuItem != null ){
            jMenuItem.doClick();
        } else {
            unproto.log.warning( "'" + value + "' initial logging level not found, using default" );
        }
    } // DstarTerminal.initMenus

    //----------------------------------------------------------------------------------------------
    /** Listen for D-Star message segments and dispatch them.
     */
    private static void listen(){
        DstarTerminal         terminal;
        DstarSerialIO.Segment segment;
        String                callsign;

        try {
            dstarSerial = new DstarSerialIO( unproto.log, localCall, serialPortId, baudRate );
            while ( true ){
                segment = dstarSerial.receive( null );
                unproto.log.finer( "===== From: " + segment.source
                                    +    "; to: " + segment.target + " =====" );
/*
//                switch ( segment.type ){
//                case DstarSerialIO.Segment.Type.CONNECT:
                    terminal = new DstarTerminal( segment.source );
                    terminal.log.info( new String( segment.body ) );
                    hashMap.put( segment.source, terminal );
//                    continue;
//                case DstarSerialIO.Segment.Type.DISCONNECT:
                    terminal = hashMap.get( segment.source );
                    if ( terminal != null ){
                        terminal.segment.setEditable( false );
                        terminal.log.info( new String( segment.body ) );
                        terminal.handler.pane.setBackground( Color.LIGHT_GRAY );
                        jTabbedPane.setTitleAt( jTabbedPane.indexOfComponent( terminal ),
                                                "[" + terminal.remoteCall + "]" );
                        hashMap.remove( terminal.remoteCall );
                        continue;
                    }
//                    break;
//                case DstarSerialIO.Segment.Type.CONNECTED_DATA:
                    terminal = hashMap.get( segment.source );
                    if ( terminal != null ){
                        terminal.log.log( Level.ALL, new String( segment.body ) );
                        continue;
                    }
//                    break;
//                }
*/
                unproto.log.finest( new String( segment.body ) );
            }
        } catch ( Exception exception ){
            unproto.log.log( Level.SEVERE, "Communications error", exception );
        }
    } // DstarTerminal.listen

    //----------------------------------------------------------------------------------------------
    /** Initialize the default window tab.
     */
    private DstarTerminal(){
        super( BoxLayout.Y_AXIS );
        final   JTextField      stationField    = new JTextField( DstarSerialIO.Segment.ADDRESS_SIZE );
                JButton         clearButton     = new JButton( "Clear" );
                Box             vBox            = new Box( BoxLayout.Y_AXIS );
                Box             hBox;

        add( vBox );
        jTabbedPane.addTab( "Unconnected", null, this, "Unconnected activity" );
        handler.setLevel( Level.CONFIG );
        this.remoteCall = null;

        vBox.add( new JPanel() );

        hBox = new Box( BoxLayout.X_AXIS );
        hBox.add( new JPanel() );
        hBox.add( new JLabel( "Destination station: " ) );
        hBox.add( stationField );
        hBox.add( new JPanel() );
        vBox.add( hBox );

        vBox.add( new JPanel() );

        hBox = new Box( BoxLayout.X_AXIS );
        hBox.add( new JPanel() );
        hBox.add( new JLabel( "Message text: " ) );
        hBox.add( messageField );
        hBox.add( new JPanel() );
        hBox.add( clearButton );
        hBox.add( new JPanel() );
        vBox.add( hBox );

        vBox.add( new JPanel() );
        vBox.setMaximumSize( vBox.getPreferredSize() );

        clearButton.addActionListener( new ActionListener(){
                public void actionPerformed( ActionEvent event ){
                    messageField.setText( null );
                }
            } ); // ActionListener implementation
        ActionListener transmitListener = new ActionListener(){
            public void actionPerformed( ActionEvent event ){
                String   remoteCall = stationField.getText().toUpperCase();
                String   text       = event.getActionCommand();
                try {
                    dstarSerial.send( new DstarSerialIO.Segment( log, localCall, remoteCall, text.getBytes() ) );
                } catch ( Exception exception ){
                    log.log( Level.SEVERE, "Communications error sending a message", exception );
                }
            }  
        }; // ActionListener implementation
        messageField.addActionListener( transmitListener );
        textFont = new Font( preferences.get(    "Window.font.name", "Monospaced" ),
                             preferences.getInt( "Window.font.style", Font.PLAIN ),
                             preferences.getInt( "Window.font.size", 14 ) );
        handler.pane.setFont( textFont );
    } // DstarTerminal constructor

    //----------------------------------------------------------------------------------------------
    /** Initialize a connected window tab.
     *
     *  @param  remoteCall              the remote station's callsign and SSID.
     */
    private DstarTerminal( final String remoteCall ){
        super( BoxLayout.Y_AXIS );
        JButton clearButton     = new JButton( "Clear" );
        Box     vBox            = new Box( BoxLayout.Y_AXIS );
        Box     hBox;

        add( vBox );
        jTabbedPane.insertTab( remoteCall, null, this, remoteCall + " activity", 0 );
        handler.setLevel( Level.INFO );
        this.remoteCall = remoteCall;

        vBox.add( new JPanel() );

        hBox = new Box( BoxLayout.X_AXIS );
        hBox.add( new JPanel() );
        hBox.add( new JLabel( "Message text:" ) );
        hBox.add( messageField );
        hBox.add( new JPanel() );
        hBox.add( clearButton );
        hBox.add( new JPanel() );
        vBox.add( hBox );

        vBox.add( new JPanel() );
        vBox.setMaximumSize( vBox.getPreferredSize() );

        clearButton.addActionListener( new ActionListener(){
                public void actionPerformed( ActionEvent event ){
                    messageField.setText( null );
                }
            } ); // ActionListener implementation
        ActionListener transmitListener = new ActionListener(){
            public void actionPerformed( ActionEvent event ){
                String text = event.getActionCommand();
                try {
                    dstarSerial.send( new DstarSerialIO.Segment( log, localCall, remoteCall, text.getBytes() ) );
                } catch ( Exception exception ){
                    log.log( Level.SEVERE, "Communications error sending a message", exception );
                }
            }  
        }; // ActionListener implementation
        messageField.addActionListener( transmitListener );
        textFont = unproto.textFont;
        handler.pane.setFont( textFont );
        repaint();
//      revalidate();
    } // DstarTerminal constructor
} // class DstarTerminal
