package com.ultimeth.util;

import java.util.*;

//==================================================================================================
/** Pair is a pair of Objects.
 *
 *  @version    2005.08.12
 *  @author     Copyright © 2002 by Dean K. Gibson.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
public class Pair <K extends Comparable <K>, V> implements Comparable <Pair <K, V>>{
    public  final K key;
    public  final V value;

    //----------------------------------------------------------------------------------------------
    /** Pair constructor.
     *
     *  @param  key                     the Comparable value.
     *  @param  value                   the associated Object.
     */
    public Pair( K key, V value ){
        this.key   = key;
        this.value = value;
    }// Pair constructor

    //----------------------------------------------------------------------------------------------
    /** Return the key's String value (Object override).
     *
     *  @return String                  - the key's String value.
     */
    public String toString(){
        return key.toString();
    }// Pair.toString

    //----------------------------------------------------------------------------------------------
    /** Compare this Pair to another Pair (Comparable &lt;Pair &lt;K, V&gt; &gt; implementation).
     *
     *  @param  pair                    the Pair &lt;K, V&gt; to compare to.
     *  @return int                     - the result of the comparison.
     */
    public int compareTo( Pair <K, V> pair ){
        return key.compareTo( pair.key );
    }// Pair.compareTo

    //----------------------------------------------------------------------------------------------
    /** Returns the comparator.
     *
     *  @return Comparator              - the comparator.
     */
    public Comparator <? extends Pair <K, V>> comparator(){
        return new Comparator <Pair <K, V>>(){
            public int compare( Pair <K, V> pair1, Pair <K, V> pair2 ){
                return pair1.compareTo( pair2 );
            }
        };
    }// Pair.comparator
}// Pair class
